@@ -89,9 +89,27 @@ def check_function(args):
     function_name = args[0]
     files = args[1:]
 
-    # Surpress output
+    # Match:
+    #  [1] start of line OR non-identifier character
+    #  [2] function name
+    #  [3] optional whitespace
+    #  [4] opening parenthesis (function call)
+    regex = rf'(^|[^[:alnum:]_]){function_name}[[:space:]]*\('
+
+    # Strip comments, then search
+    cmd = (
+        "sed -E "
+        "'"
+        "s://.*$::g; " # C/C++ single-line comments
+        "s:#.*$::g; " # Python single-line comments
+        ":a; /\\/\\*/{N; s:/\\*.*?\\*/::g; ba}"
+        "' "
+        + " ".join(files)
+        + f" | grep -E '{regex}'"
+    )
+
     function_popen = subprocess.Popen(
-        ["grep", f"[^[:alpha:]]{function_name}[[:space:]]*("] + files,
+        ["bash", "-c", cmd],
         stdout=subprocess.PIPE,
         stderr=subprocess.PIPE,
     )
@@ -102,6 +120,101 @@ def check_function(args):
     else:
         print(f"Used {function_name} PASSED")
 
+def check_object(args):
+    """Will be followed by an object and a list of files to ensure that the function is
+    used by one of those files. 
+    
+    This primarily is used for C++ objects from instances of the ifstream and ofstream operators
+    
+    Arguments:
+        object_name: the object name to check files for usage of
+        *files: the files to check for the function name
+
+    Returns:
+        None
+    """
+    
+    check_test()
+    args = args.split()
+    object_name = args[0]
+    files = args[1:]
+
+    # Match:
+    #  [1] start of line OR non-identifier character
+    #  [2] object name
+    #  [3] optional whitespace
+    #  [4] stream operator << or >>
+    regex = rf'(^|[^[:alnum:]_]){object_name}[[:space:]]*(<<|>>)'
+
+    # Strip comments, then search
+    cmd = (
+        "sed -E "
+        "'s://.*$::g; "              # remove // comments
+        ":a; /\\/\\*/{N; s:/\\*.*?\\*/::g; ba}' "
+        + " ".join(files)
+        + f" | grep -E '{regex}'"
+    )
+
+    # Surpress output
+    function_popen = subprocess.Popen(
+        ["bash", "-c", cmd],
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE,
+    )
+
+    function_popen.communicate()
+    if function_popen.returncode:
+        print(f"Not using {object_name} FAILED")
+    else:
+        print(f"Used {object_name} PASSED")
+        
+def check_container(args):
+    """Will be followed by a container and a list of files to ensure that the function is
+    used by one of those files. 
+    
+    This primarily is used for C++ containers such as vector<double> or std::vector v;
+    
+    Arguments:
+        object_name: the object name to check files for usage of
+        *files: the files to check for the function name
+
+    Returns:
+        None
+    """
+    
+    check_test()
+    args = args.split()
+    container_name = args[0]
+    files = args[1:]
+    
+    # Match:
+    #  [1] start of line OR non-identifier character
+    #  [2] container name
+    #  [3] optional whitespace
+    #  [4] either '<' (template usage) or space (e.g. std::vector v;)
+    regex = rf'(^|[^[:alnum:]_]){container_name}[[:space:]]*(<|[[:space:]])'
+
+    # Strip comments, then search
+    cmd = (
+        "sed -E "
+        "'s://.*$::g; "
+        ":a; /\\/\\*/{N; s:/\\*.*?\\*/::g; ba}' "
+        + " ".join(files)
+        + f" | grep -E '{regex}'"
+    )
+
+    function_popen = subprocess.Popen(
+        ["bash", "-c", cmd],
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE,
+    )
+
+    function_popen.communicate()
+    if function_popen.returncode:
+        print(f"Not using {container_name} FAILED")
+    else:
+        print(f"Used {container_name} PASSED")
+    
 
 def check_not_function(args):
     """Will be followed by a function name and a list of files to check to ensure that the function
@@ -184,18 +297,16 @@ def run_command_noerror(command):
         print("PASSED")
 
 
-def compare(file1, file2):
-    """Will be followed by two files to compare.
+def compare(files):
+    """Will be followed by files to compare.
 
     Arguments:
-        file1: The first file to compare
-        file2: The second file to compare
+        files: A string of space-separated files to compare
 
     Returns:
         None
     """
-    cmps.append(file1)
-    cmps.append(file2)
+    cmps.append(files.split())
 
 
 def test_case(test_case_command):
@@ -416,6 +527,8 @@ this in the function itself.
 tag_func_map = {
     "C": compile_code,
     "CF": check_function,
+    "CO" : check_object,
+    "CC": check_container,
     "NCF": check_not_function,
     "CMD": run_command,
     "TCMD": run_command_noerror,
@@ -449,6 +562,12 @@ def setup():
     for file in files:
         open(file, "w").close()
 
+    # Reset test case variables
+    global expected_exit_code
+    expected_exit_code = -1
+    global cmps
+    cmps = []
+
 
 def parse_tags(tags: list[str]):
     """Given list of strings, parses and executes tags
@@ -459,26 +578,101 @@ def parse_tags(tags: list[str]):
     Returns:
         None
     """
-    tag_pattern = r"([A-Z]+) (.*)"
-    for tag_line in tags:
-        tag_match = re.match(tag_pattern, tag_line)
+    # Pattern matches: TAG arguments (arguments required)
+    tag_pattern = r"([A-Z_]+)\s+(.*)"
+    # Pattern for tag with optional arguments
+    tag_only_pattern = r"([A-Z_]+)\s*$"
 
-        # If line does not match tag format
+    valid_tags = set(tag_func_map.keys())
+
+    # Track if we're inside a CRT_HW block (content to ignore)
+    in_crt_hw_block = False
+
+    for line_num, tag_line in enumerate(tags, start=1):
+        # Skip empty lines and comments
+        stripped = tag_line.strip()
+        if not stripped or stripped.startswith("#"):
+            continue
+
+        # Check for CRT_HW block delimiter
+        if stripped.startswith("CRT_HW"):
+            in_crt_hw_block = not in_crt_hw_block
+            continue
+
+        # Skip lines inside CRT_HW block
+        if in_crt_hw_block:
+            continue
+
+        tag_match = re.match(tag_pattern, tag_line)
+        tag_only_match = re.match(tag_only_pattern, tag_line)
+
+        # Check for tag without arguments
+        if tag_only_match and not tag_match:
+            tag = tag_only_match.group(1)
+            if tag in valid_tags:
+                print(f"Error on line {line_num}: Tag '{tag}' requires arguments")
+                print(f"  {line_num}: {tag_line.rstrip()}")
+                sys.exit(1)
+            elif tag.isupper():
+                print(f"Error on line {line_num}: Unknown tag '{tag}'")
+                print(f"  {line_num}: {tag_line.rstrip()}")
+                print(f"  Valid tags are: {', '.join(sorted(valid_tags))}")
+                sys.exit(1)
+            continue
+
+        # If line does not match tag format, skip non-tag lines
         if not tag_match:
+            # Check if it looks like an invalid tag (starts with uppercase letters)
+            potential_tag = re.match(r"([A-Z]+)", tag_line)
+            if potential_tag:
+                tag = potential_tag.group(1)
+                if tag not in valid_tags and len(tag) <= 4:
+                    print(f"Error on line {line_num}: Unknown tag '{tag}'")
+                    print(f"  {line_num}: {tag_line.rstrip()}")
+                    print(f"  Valid tags are: {', '.join(sorted(valid_tags))}")
+                    sys.exit(1)
             continue
 
         tag = tag_match.group(1)
         args = tag_match.group(2)
 
+        # Check for unknown tag
+        if tag not in valid_tags:
+            print(f"Error on line {line_num}: Unknown tag '{tag}'")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            print(f"  Valid tags are: {', '.join(sorted(valid_tags))}")
+            sys.exit(1)
+
+        # Check for empty arguments
+        if not args.strip():
+            print(f"Error on line {line_num}: Tag '{tag}' requires arguments")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            sys.exit(1)
+
         # Execute function based on tag-function mapping
         try:
             tag_func_map[tag](args)
-        except KeyError:
-            # Tag was not found in dictionary
-            continue
-        except (TypeError, ValueError):
+        except TypeError as e:
+            print(f"Error on line {line_num}: Invalid arguments for tag '{tag}'")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            print(f"  Details: {e}")
+            sys.exit(1)
+        except ValueError as e:
+            print(f"Error on line {line_num}: Invalid value for tag '{tag}'")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            print(f"  Details: {e}")
+            sys.exit(1)
+        except FileNotFoundError as e:
+            print(f"Error on line {line_num}: File not found for tag '{tag}'")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            print(f"  Details: {e}")
+            sys.exit(1)
+        except Exception as e:
+            print(f"Error on line {line_num}: Unexpected error for tag '{tag}'")
+            print(f"  {line_num}: {tag_line.rstrip()}")
+            print(f"  Details: {e}")
             traceback.print_exc()
-            print(f"Invalid arguments for tag {tag} {args}")
+            sys.exit(1)
 
 
 def parse_diff(diff_lines: list[str]):
@@ -575,7 +769,7 @@ def check_test():
 
     # Compare files handling, do not surpress output
     for files in cmps:
-        cmd_popen = subprocess.Popen(["cmp", files])
+        cmd_popen = subprocess.Popen(["cmp"] + files)
         cmd_popen.communicate()
         if cmd_popen.returncode:
             passed = False
@@ -660,7 +854,14 @@ def run_evaluation(codeval_file):
     global test_case_total
     with open(codeval_file, "r") as infile:
         testcases = infile.readlines()
+        in_crt_hw_block = False
         for testcase in testcases:
+            stripped = testcase.strip()
+            if stripped.startswith("CRT_HW"):
+                in_crt_hw_block = not in_crt_hw_block
+                continue
+            if in_crt_hw_block:
+                continue
             parts = testcase.split(" ", 1)
             tag = parts[0]
             if tag == "T" or tag == "HT" or tag == "TCMD":
